% file jed_grep.sl
%
% A complex function wich tries to get a file to open from the current line in
% current buffer. As the names reveal, this was created to follow grep-output
% lines. Now does a lot more, and can recognise #include directives, grep-like
% output (with and without line numbers, and if everything else fails, a
% simple file name.
% Uses walk.sl, so to turn back press the key bound to walk_backward()
% (or to ntags_back())
% this was written a lot of time ago, when I was starting to learn slang, so
% it looks very ugly.
%
% a powerful example of how to use this is doing:
% fgrep -n "sometext" * | jed
% go to a line looking promising and press the Ctrl-V key-sequence (or
% whatever you bound follow_grep() to).
%
% if you have grep_mode in your autoload list, you could even do
%   fgrep -n "sometext" * | jed -f followgrep_mode
% and simply double-click or press Enter at the line in question.
% or you bind "grep_and_follow" to a key and do the grep inside jed.
%
% You can customize the follogrep_mode by the use of "followgrep_mode_hook".
% E.g. if you have defined a funciton that closes the buffer and the  window
% it was in (if there was more than one window), the following snippet in
% your setup file (.jedrc, say) would set Q to call this function:
%
% define followgrep_mode_hook ()
% {
%    local_setkey ("close_buffer", "q");
% }
%



%_debug_info = 1;
%_traceback=1;
%
implements("FollowGrep");

require("walk");

custom_variable("FollowGrep_Dirs", ".");
custom_variable("Include_Braket_Dirs", "/usr/include,.");
custom_variable("Include_Quote_Dirs", ".,./include,../include,./source,../source,./src,../src");

define follow_file()
{
        variable wrongchars = "^\t\n ()[]{}\"\'\$;:,";
        variable filename;
        variable found = -1;

        push_mark();        % poppato da pop_mark(1)
        bskip_chars(wrongchars);
        push_mark();        % poppato da bufsubstr()
        skip_chars(wrongchars);
        filename = bufsubstr();
        if (strlen(filename) == 0)
                return found;
        filename = expand_filename(filename);
        pop_mark(1);
        if (stat_file(filename) != NULL) {
                () = find_file(filename);
                walk_store_marked_position();
                found = 0;
        }
        return found;
}

define follow_list()
{
        variable filename;
        variable found = -1;

        push_mark();        % poppato da pop_mark(1)
        bol();
        push_mark();        % poppato da bufsubstr()
        skip_chars("^\t\n");
        filename = expand_filename(bufsubstr());
        pop_mark(1);
        if (stat_file(filename) != NULL) {
                () = find_file(filename);
                walk_store_marked_position();
                found = 0;
        } else
                found = follow_file();
        return found;
}

define find_include_match(match, Incl_Dirs)
{
        variable file, FG_Path;
        variable found = -1;

        () = right(1);
        push_mark();
        () = ffind(match);
        file = bufsubstr();

        foreach (strchop(Incl_Dirs, ',', 0)) {
                FG_Path = dircat((), file);
                if (file_status(FG_Path) == 1) {
                        () = find_file(FG_Path);
                        walk_store_marked_position();
                        found = 0;
                        break;
                }
        }
        return found;
}

define follow_include()
{
        variable size = 0;
        variable found = -1;

        push_mark();
        bol();

        !if (looking_at_char('#'))
                skip_chars("^#\n");

        if (looking_at_char('#'))
                size = ffind("include");

        if (size > 0) {
                () = right(size);

                if (ffind("<"))
                        found = find_include_match(">", Include_Braket_Dirs);
                else if (ffind("\""))
                        found = find_include_match("\"", Include_Quote_Dirs);
        }

        pop_mark(1);
        return found;
}

public define follow_grep()
{
        variable name;
        variable numline;
        variable FG_Path;
        variable found = 0;

        % Save Current Pos
        walk_mark_current_position();

        push_mark();        % poppato da pop_mark(1)
        bol();
        push_mark();        % poppato da bufsubstr()
        skip_chars("^:\n");

        if (looking_at_char(':')) {
                name = bufsubstr();
                () = right(1);
                push_mark();       % poppato da bufsubstr()
                skip_chars("[0-9]");
                numline = integer(bufsubstr());

                foreach (strchop(FollowGrep_Dirs, ',', 0)) {
                        FG_Path = dircat((), name);
                        if (file_status(FG_Path) == 1) {
                                found = 1;
                                break;
                        }
                }
        } else
                pop_mark(0);

        pop_mark(1);
        if (found == 1) {
                () = find_file(FG_Path);
                goto_line(numline);
                goto_column(0);
                % Now I can Store the old pos permanently.
                walk_store_marked_position();
        } else
                if (follow_list() == -1)
                        if (follow_include() == -1)
                                find_tag();
}

% appendix: invoce grep from within jed and store the result in a
% dedicated buffer with appropriate mode


% --- some helpers ---------------------------------------------

%!%+
%\function{custom}
%\synopsis{make variable initialization user-customizable}
%\usage{AnyType custom (String var, AnyType default)}
%\description
%   If the variable c_var is defined, it is returned, otherwise
%   default is returned. Use this instead of custom_variable, if you
%   want to change the variable but preserve the custom value for a
%   possible reset.
%\seealso{custom_variable}
%\example
%  set TAB to 4 or custom value:
%#v+
%    TAB = custom("ChartableTabSpacing", 4)
%#v-
%\notes
%   this function is hopefully some day part of site.sl
%!%-
define custom (varname, default)
{
   variable r = __get_reference (varname);
   if (r != NULL)
     if (__is_initialized(r))
     return(@r);
   return(default);
}

%!%+
%\function{rename_buffer_overwrite}
%\synopsis{rename the buffer, overwriting an existing buffer}
%\usage{ Void rename_buffer_overwrite (buffer_name)}
%\description
%   renames the active buffer to buffer_name, if a buffer of this name
%   exists, it is deleted first
%\seealso{rename_buffer}
%!%-
define rename_buffer_overwrite (buffer_name)
{
   if (bufferp(buffer_name))
     delbuf(buffer_name);
   rename_buffer(buffer_name);
}


%---- now define the stuff ----------------------------

require("keydefs");

% Variables:

custom_variable ("JedGrepCommand", "grep -n"); % the grep command
public variable LAST_GREP = "";                % similar to LAST_SEARCH

create_syntax_table("FollowGrep");
!if (keymap_p ("FollowGrep")) make_keymap ("FollowGrep");

definekey ("follow_grep",          "^M",    "FollowGrep");  % Return
definekey ("delbuf(whatbuf)",      "q",     "FollowGrep");
definekey ("delbuf(whatbuf)",      "\d155", "FollowGrep");  % "meta-escape"
definekey ("followgrep_remove_line_or_region", "d", "FollowGrep");

#ifdef HAS_DFA_SYNTAX
%%% DFA_CACHE_BEGIN %%%
static define setup_dfa_callback (name)
{
        dfa_enable_highlight_cache("jed_grep.dfa", name);
        dfa_define_highlight_rule("^[^:]*:", "keyword", name);
        dfa_define_highlight_rule("[0-9]+:", "number", name);
        % Uhm, this matches numbers followed by ':' anywhere. If this really
        % annoys you, either:
        % 1 - disable dfa highlighting
        % 2 - comment the two dfa_define_highlight_rule above an use instead:
%       dfa_define_highlight_rule("^[^:]*:[0-9]+:", "keyword", name);

   dfa_build_highlight_table(name);
}
dfa_set_init_callback (&setup_dfa_callback, "FollowGrep");
%%% DFA_CACHE_END %%%
enable_dfa_syntax_for_mode("FollowGrep");
#endif

% Mouse bindings
define followgrep_mouse_2click_hook (line, col, but, shift)
{
   follow_grep();
   return (0);
}

% Update hook to move line mark.
static variable FGLine = NULL;
static define fg_update_hook()
{
   FGLine = create_line_mark(color_number("menu_selection"));
}

% a mode dedicated to the follow_grep command
public define followgrep_mode ()
{
   rename_buffer_overwrite("*grep_output*");
   set_buffer_modified_flag (0); % so delbuf doesnot ask whether to save first
   set_readonly(1);
   set_mode("FollowGrep", 0);
   use_keymap ("FollowGrep");
   use_syntax_table ("FollowGrep");
   set_buffer_hook ( "mouse_2click", &followgrep_mouse_2click_hook);
   set_buffer_hook("update_hook", &fg_update_hook);
   run_mode_hooks("followgrep_mode_hook");
}

% Main function
public define grep_and_follow ()
{
   variable searchdir;
   variable searchfile;
   variable search_where;
   variable allfiles = "";

#ifdef UNIX
   allfiles = "*";
#elifdef IBMPC_SYSTEM
   allfiles = "*.*";
#endif
% TODO: What should this be on VMS?

   LAST_GREP = read_mini("(Flags and) String to grep: ", LAST_GREP, "");

   searchdir = read_with_completion ("Where to grep: ", "", allfiles, 'f');

   searchfile = path_basename(searchdir);
   if (searchfile == "")
     searchfile = allfiles;
   searchdir = path_dirname(searchdir);
   search_where = path_concat(searchdir, searchfile);

   shell_perform_cmd(JedGrepCommand+" "+LAST_GREP+" "+search_where, 0);

   bob;
   insert("Command: "+JedGrepCommand+" "+LAST_GREP+" "+ search_where + "\n");
   followgrep_mode;
}

public define followgrep_remove_line_or_region()
{
        set_readonly(0);
        if (markp()) {
                check_region(0);
                narrow ();
                bob();
                while (not eobp())
                        call("kill_line");
                widen ();
                call("kill_line");
        } else {
                bol();
                call("kill_line");
        }
        set_readonly(1);
}

provide("jed_grep");
